#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>

#include <Rinternals.h>
#include "R_ext/PrtUtil.h"
#include "R_ext/eventloop.h"
#include "ipcevent.h"
#include "ipctk.h"

/* From Parse.h -- must find better solution: */
#define PARSE_NULL              0
#define PARSE_OK                1
#define PARSE_INCOMPLETE        2
#define PARSE_ERROR             3
#define PARSE_EOF               4

SEXP R_ParseVector(SEXP, int, int *);

#include "tcltk.h" /* declarations of our `public' interface */


ipc_event_sink_t sink;


static void handler(ipc_event_t event, ipc_bool_t want_reply)
{
    ipctk_call_header_t hdr;
    ipctk_result_code_t code;
    ipc_size_t size;
    char *data, *msg;

    IPC_GetEventData(event, (void **) &data, &size);
    hdr = *((ipctk_call_header_t *) data);
    data += sizeof(hdr);

    switch (hdr.type) {
    case IPCTK_EVAL:
	{
	    int status, i;
	    SEXP text, expr, ans;

	    text = PROTECT(allocVector(STRSXP, hdr.argc - 1));
	    data += strlen(data) + 1;
	    for (i = 0 ; i < hdr.argc - 1; i++) {
		SET_STRING_ELT(text, i, mkChar(data));
		data += strlen(data) + 1;
	    }

	    expr = PROTECT(R_ParseVector(text, -1, &status));
	    if (status != PARSE_OK) {
		code = IPCTK_ERROR;
		msg = "parse error in R expression";
	    }
	    else {
		/* Note that expr becomes an EXPRSXP and hence we need
		   the loop below (a straight eval(expr, R_GlobalEnv)
		   won't work) */
		int n = length(expr);
		for(i = 0 ; i < n ; i++)
		    /******** need an exit trap here */
		    ans = eval(VECTOR_ELT(expr, i), R_GlobalEnv);
		code = IPCTK_OK;
		msg = "";
	    }
	    UNPROTECT(2);
	    break;
	}
    case IPCTK_CALL:
	{
	    int i;
	    SEXP expr, ans, fun, last;

	    expr = R_NilValue;
	    data += strlen(data) + 1;
	    fun = (SEXP) strtoul(data, NULL, 16); /**** make a stable pointer*/
	    data += strlen(data) + 1;

	    PROTECT(expr = LCONS(fun, R_NilValue));
	    last = expr;
	    for (i = 0; i < hdr.argc - 2; i++) {
		SETCDR(last, LCONS(mkString(data), R_NilValue));
		data += strlen(data) + 1;
		last = CDR(last);
	    }
	    UNPROTECT(1);

	    PROTECT(expr = LCONS(install("try"), LCONS(expr, R_NilValue)));
	    /******** need an exit trap here */
	    ans = eval(expr, R_GlobalEnv);
	    UNPROTECT(1);

	    code = IPCTK_OK;
	    msg = "";
	    break;
	}
    default:
	code = IPCTK_ERROR;
	msg = "unknown callback type";
    }
    IPC_PutEventData(event, &code, sizeof(code));
    if (code == IPCTK_ERROR)
	IPC_AppendEventData(event, msg, strlen(msg) + 1);
}
#ifdef DODO
static int R_call(ClientData clientData,
		  Tcl_Interp *interp,
		  int argc,
		  char *argv[])
{
>>>>>>>>>>>>>
    int i;
    SEXP expr, ans, fun, alist;

    alist = R_NilValue;
    for (i = argc - 1 ; i > 1 ; i--){
	PROTECT(alist);
	alist = LCONS(mkString(argv[i]), alist);
	UNPROTECT(1);
    }

    fun = (SEXP) strtoul(argv[1], NULL, 16);

    expr = LCONS(fun, alist);
    expr = LCONS(install("try"), LCONS(expr, R_NilValue));

    ans = eval(expr, R_GlobalEnv);

    return TCL_OK;
}
#endif

SEXP tk_eval(char *cmd)
{
    ipc_event_t event;
    ipctk_result_code_t code;
    size_t size = strlen(cmd);
    char *data;
    SEXP ans;

    if (IPC_CreateEvent(&event) != 0)
	error("event creation failed");
    if (IPC_PutEventData(event, cmd, size + 1) != 0) {
	IPC_DisposeEvent(event);
	error("adding event data failed");
    }
    if (IPC_SendEvent(sink, event, 1) != 0) {
	IPC_DisposeEvent(event);
	error("send failed");
    }
    if (IPC_GetEventData(event, (void **) &data, &size) != 0) {
	IPC_DisposeEvent(event);
	error("reading event data failed");
    }

    code = *((ipctk_result_code_t *) data);
    data += sizeof(code);
    if (code == IPCTK_ERROR) {
	char p[512];
	if (strlen(data) > 500)
	    strcpy(p, "tcl error.\n");
	else
	    sprintf(p, "[tcl] %s.\n", data);
	IPC_DisposeEvent(event);
	error(p);
    }
    ans = mkString(data);
    IPC_DisposeEvent(event);
    return ans;
}

SEXP dotTcl(SEXP args)
{
    if(!isValidString(CADR(args)))
	error("invalid argument");
    return tk_eval(CHAR(STRING_ELT(CADR(args), 0)));
}

/* FIXME get rid of fixed size buffers in do_Tclcallback */
SEXP dotTclcallback(SEXP args)
{
    SEXP ans, closure = CADR(args), formals;

    char buf[256], tmp[20];

    if (!isFunction(closure))
    	error("argument is not a function");

    formals = FORMALS(closure);

    sprintf(buf, "{ R_call 0x%lx", (unsigned long) closure);

    /* FIXME: we really should do something to ensure that "closure"
       is protected from later GCs. Otherwise we'll have buttons that
       make R go Boom... We need register_callback(closure) or
       something to that effect + a run through registered callbacks
       in the MarkPhase of GC. [Currently, this is handled by having
       .Tcl.args assign an alias to the function in the environment of
       the window with which the callback is associated] */

    while ( formals != R_NilValue )
    {
	if (TAG(formals) ==  R_DotsSymbol) break;
	sprintf(tmp, " %%%s", CHAR(PRINTNAME(TAG(formals))));
	strcat(buf, tmp);
	formals = CDR(formals);
    }
    strcat(buf, " }");
    ans = mkString(buf);
    return ans;
}

static void (* OldHandler)(void);
static int OldTimeout;
static int Tcl_loaded = 0;

void TclHandler(void)
{
    int fd;
    IPC_SinkFD(sink, &fd);

    for (;;) {
	fd_set rfds, wfds, efds;
	struct timeval tv;

	tv.tv_sec = 0;
	tv.tv_usec = 0;

	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	FD_ZERO(&efds);
	FD_SET(fd, &rfds);

	if (select(fd + 1, &rfds, &wfds, &efds, &tv) > 0 &&
	    FD_ISSET(fd, &rfds)) {
	    if (IPC_HandleEvent(sink) != 0) {
		fprintf(stderr, "event handler failed\n");
		break;
	    }
	}
	else
	    break;
    }

    OldHandler();
}

void addTcl(void)
{
    if (Tcl_loaded)
	error("Tcl already loaded");
    Tcl_loaded = 1;
    OldHandler = R_PolledEvents;
    OldTimeout = R_wait_usec;
    R_PolledEvents = TclHandler;
    if ( R_wait_usec > 10000 || R_wait_usec == 0)
	R_wait_usec = 10000;
}

void delTcl(void)
{
    if (!Tcl_loaded)
	error("Tcl is not loaded");
    if (R_PolledEvents != TclHandler)
	error("Tcl is not last loaded handler");
    R_PolledEvents = OldHandler;
    R_wait_usec = OldTimeout;
    Tcl_loaded = 0;
}

void tcltk_init(char **path)
{
    char *argv[4];

    argv[0] = "xyzq";
    argv[1] = "R_eval";
    argv[2] = "R_call";
    argv[3] = NULL;

    if (IPC_StartServer(*path, argv, &sink, handler) != 0)
	error("connect failed\n");
    addTcl();
}
